'----------------------------------------------------------------------
'
' Copyright 2015 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------

' This example demonstrates how to use the WiFi and network routines to request
' data from a remote HTTP host.  The Main() routine is structured as a state
' machine that performs various operations in each of several states.  As an
' aid to understanding the operation, much information is output during the
' process.

Option ConsoleSpeed 230400
Option UserInit myInit
Option UserPostInit myPostInit
Option Include IPAddress_t

' set the SSID and password of the desired WiFi access point
#if Not defined(ssid)
Const ssid as String = "mySSID"
Const passwd as String = "myPasswd"
#endif

' set the domain and URL strings here
Const domain as String = "www.zbasic.net"
Const request as String = "GET /ZBasicTest/testPage.html HTTP/1.0" & CRLF

Const CRLF as String = Chr(&H0d) & Chr(&H0a)

Const FRC2_ticksPerSec as UnsignedLong = 5000000
Const FRC2_replyWaitTime as UnsignedLong = FRC2_ticksPerSec * 20

Const WifiConnected as Byte = 5

' define the various states
Enum NetState
	AwaitingWifiConnection
	ConnectToHost
	SendHostRequest
	AwaitingHostReply
	DisconnectFromHost
	NetOperationCompleted
End Enum

Dim state as NetState
Dim count as Integer
Dim haveData as Boolean
Dim hand as Integer
Dim hostAddr as UnsignedLong
Dim hostReply(1 to 500) as Byte
Dim hostReplyLen as UnsignedInteger
Dim connected as Boolean
Dim lastTick as UnsignedLong
Dim startTick as UnsignedLong
Dim lastState as NetState

Sub Main()
	Dim stat as Integer
	Dim curTick as UnsignedLong
	Dim report as Boolean
	Dim wifiStatus as Byte
	Dim ipAddr as UnsignedLong

	' get the current 200nS tick count and WiFi status
	curTick = Register.FRC2_COUNT
	wifiStatus = Wifi.Status()
	report = False

	' determine if a report should be made
	If (state <> NetOperationCompleted) Then
		' check for 1 second having elapsed
		If ((curTick - lastTick) > FRC2_ticksPerSec) Then
			report = True
			lastTick = curTick
			count = count + 1
		End If

		' check for a state change
		If (state <> lastState) Then
			report = True
			lastState = state
		End If

		' check for WiFi connection
		If ((state = AwaitingWifiConnection) And (wifiStatus = WifiConnected)) Then
			report = True
		End If
	End If

	' get the current WiFi status, report state information
	If (report) Then
		Debug.Print "count = "; count;
		Debug.Print ", state = "; state;
		Debug.Print ", WiFi status is "; wifiStatus;
		If (hand >= 0) Then
			Debug.Print ", NetStatus() is 0x"; CStrHex(Net.Status(hand));
		End If
		Debug.Print
	End If

	Select Case state
	Case AwaitingWifiConnection
		' see if an AP connection has been established
		If (wifiStatus = WifiConnected) Then
			ipAddr = Wifi.GetIP(0)
			Call dspIP("connected to access point, IP = ", ipAddr)
			Debug.Print

			hand = -1
			hostReplyLen = 0
			haveData = False
			If (ipAddr <> 0) Then
				hand = Net.Open()
			End If
			If (hand >= 0) Then
				' set up the network callback, look up host IP address
				stat = Net.SetCallback(hand, @netCallback)
				Debug.Print "Net.SetCallback() returns "; stat

				Debug.Print "looking up domain name "; domain
				hostAddr = Net.GetHostByName(domain)
				Call dspIP("the domain name resolves to ", hostAddr)
				Debug.Print
				state = ConnectToHost
			Else
				state = NetOperationCompleted
			End If
		End If

	Case ConnectToHost
		If ((hand >= 0) And (hostAddr <> 0)) Then
			' connect to the host on port 80
			Dim port as Integer
			Debug.Print "connecting..."
			stat = Net.Connect(hand, hostAddr, 80, 0)
			Debug.Print "Net.Connect() returns "; stat; ", NetStatus() is 0x"; CStrHex(Net.Status(hand))
			stat = Net.RemoteHost(hand, ipAddr, port)
			Call dspIP("the remote host IP is ", ipAddr)
			Debug.Print ", port "; port
			state = SendHostRequest
		Else
			state = NetOperationCompleted
		End If

	Case SendHostRequest
		If (hand >= 0) Then
			' send page request to host
			stat = Net.SendStr(hand, request & "host: " & domain & (CRLF & CRLF))
			Debug.Print "Net.SendStr() returns "; stat
			state = AwaitingHostReply
			startTick = curTick
		Else
			state = NetOperationCompleted
		End If

	Case AwaitingHostReply
		If (haveData) Then
			If (hostReplyLen > 0) Then
				Dim i as UnsignedInteger
				Debug.Print
				Debug.Print "Host reply:"
				Debug.Print "----------------------------------"
				For i = 1 to hostReplyLen
					If (hostReply(i) = &H0a) Then
						Debug.Print
					ElseIf (hostReply(i) <> &H0d) Then
						Debug.Print Chr(hostReply(i));
					End If
				Next i
				Debug.Print "----------------------------------"
				Debug.Print
			End If
			haveData = false
			state = DisconnectFromHost
		ElseIf ((curTick - startTick) > FRC2_replyWaitTime) Then
			Debug.Print "Timed out waiting for host reply"
			state = NetOperationCompleted
		End If

	Case DisconnectFromHost
		If (hand >= 0) Then
			stat = Net.Disconnect(hand)
			Debug.Print "Net.Disconnect() returns "; stat
			stat = Net.Close(hand)
			Debug.Print "Net.Close() returns "; stat
			hand = -1
		End If
		state = NetOperationCompleted

	End Select
End Sub

Sub myPostInit()
	Debug.Print
	hand = -1
	state = AwaitingWifiConnection
	lastState = state
	lastTick = Register.FRC2_COUNT
End Sub

Sub myInit()
	Call Wifi.SetMode(1)
	Call Wifi.SetConfig(ssid, passwd)
End Sub

'------------------------------------------------------------------------------

'
'' netCallback
'
' This is a callback routine for network operations.
'
Sub netCallback(ByVal cbtype as Byte, ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	Debug.Print "in netCallback(), chan "; handle; " ";
	Select Case cbtype
	Case 0
		' connected to remote host
		Debug.Print "cb: connect"
	Case 1
		' disconnected from remote host
		Debug.Print "cb: disconnect"
	Case 2
		' data received from remote host
		Debug.Print "cb: receive ("; cnt; " bytes)"
		If (cnt > 0) Then
			If (cnt > SizeOf(hostReply)) Then
				cnt = SizeOf(hostReply)
			End If
			Call MemCopy(@hostReply, @data, cnt)
			hostReplyLen = cnt
			haveData = True
		End If
	Case 3
		' sent data to remote host
		Debug.Print "cb: send"
		stat = Net.Close(handle)
	Case 4
		' write to remote host completed
		Debug.Print "cb: write finish"
	Case 5
		' reconnected to remote host
		Debug.Print "cb: reconnect"
	Case Else
		' unknown callback type
		Debug.Print "cb: unknown ("; cbtype; ")"
	End Select
End Sub

'------------------------------------------------------------------------------

'
'' dspIP
'
' This routine outputs an IP address in "dotted quad" notation preceded
' by an optional caption.
'
Sub dspIP(ByVal caption as String, ByVal ip as UnsignedLong)
	Dim i as Integer
	Dim ipAddr as IPAddress_t

	ipAddr.w = ip
	Debug.Print caption;
	For i = 1 to 4
		Debug.Print ipAddr.b(i);
		If (i < 4) Then
			Debug.Print ".";
		End If
	Next i
End Sub

