'----------------------------------------------------------------------
'
' Copyright 2015 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------

' This example demonstrates how to use the WiFi and network routines to request
' data from a remote NTP time server using the UDP protocol.  The Main() routine
' is structured as a state machine that performs various operations in each of
' several states.  As an aid to understanding the operation, much information is
' output during the process.

Option ConsoleSpeed 230400
Option UserInit myInit
Option UserPostInit myPostInit
Option Include IPAddress_t

' set the SSID and password of the WiFi access point
#if Not defined(ssid)
Const ssid as String = "mySSID"
Const passwd as String = "myPasswd"
#endif

#define NTP_CLIENT

' uncomment the following line to specify the NTP server by name, otherwise the IP address will be used
#define USE_NTP_SERVER_NAME

' set the NTP server name and/or IP address
Const NtpServerName as String = "pool.ntp.org"
Const NtpServerIP as UnsignedLong = Net.Address(192, 168, 0, 1)

Const ntpDataSize as UnsignedInteger = 48
Const NtpPort as Integer = 123

Const FRC2_ticksPerSec as UnsignedLong = 5000000
Const FRC2_replyWaitTime as UnsignedLong = FRC2_ticksPerSec * 20

Const WifiConnected as Byte = 5

' define a union for extracting the NTP data
Union NtpData_t
	Dim d as UnsignedLong
	Dim b(1 to 4) as Byte
End Union

' define the various states
Enum NetState
	AwaitingWifiConnection
	ConnectToHost
	SendHostRequest
	AwaitingHostReply
	DisconnectFromHost
	NetOperationCompleted
End Enum

Dim state as NetState
Dim count as Integer
Dim haveData as Boolean
Dim hand as Integer
Dim hostAddr as UnsignedLong
Dim lastTick as UnsignedLong
Dim startTick as UnsignedLong
Dim lastState as NetState
#if defined(NTP_CLIENT)
Dim ntpTime as UnsignedLong
#endif

Sub Main()
	Dim stat as Integer
	Dim curTick as UnsignedLong
	Dim report as Boolean
	Dim wifiStatus as Byte
	Dim ipAddr as UnsignedLong

	' get the current 200nS tick count and WiFi status
	curTick = Register.FRC2_COUNT
	wifiStatus = Wifi.Status()
	report = False

	' determine if a report should be made
	If (state <> NetOperationCompleted) Then
		' check for 1 second having elapsed
		If ((curTick - lastTick) > FRC2_ticksPerSec) Then
			report = True
			lastTick = curTick
			count = count + 1
		End If

		' check for a state change
		If (state <> lastState) Then
			report = True
			lastState = state
		End If

		' check for WiFi connection
		If ((state = AwaitingWifiConnection) And (wifiStatus = WifiConnected)) Then
			report = True
		End If
	End If

	' get the current WiFi status, report state information
	If (report) Then
		Debug.Print "count = "; count;
		Debug.Print ", state = "; state;
		Debug.Print ", WiFi status is "; wifiStatus;
		If (hand >= 0) Then
			Debug.Print ", NetStatus() is 0x"; CStrHex(Net.Status(hand));
		End If
		Debug.Print
	End If

	Select Case state
	Case AwaitingWifiConnection
		' see if an AP connection has been established
		If (wifiStatus = WifiConnected) Then
			ipAddr = Wifi.GetIP(0)
			Call dspIP("connected to access point, IP = ", ipAddr)
			Debug.Print

			hand = -1
			haveData = False
			If (ipAddr <> 0) Then
				hand = Net.Open()
			End If
			If (hand >= 0) Then
				' set up the network callback, look up host IP address
				stat = Net.SetCallback(hand, @netCallback)
				Debug.Print "Net.SetCallback() returns "; stat

#if defined(USE_NTP_SERVER_NAME)
				Debug.Print "looking up server name "; NtpServerName
				hostAddr = Net.GetHostByName(NtpServerName)
				Call dspIP("the NTP server name name resolves to ", hostAddr)
#else
				hostAddr = NtpServerIP
				Call dspIP("the NTP server address is ", hostAddr)
#endif
				Debug.Print
				state = ConnectToHost
			Else
				state = NetOperationCompleted
			End If
		End If

	Case ConnectToHost
		If ((hand >= 0) And (hostAddr <> 0)) Then

			stat = Net.Connect(hand, hostAddr, NtpPort, 1)
			Debug.Print "Net.Connect() returns "; stat
			state = SendHostRequest
		Else
			state = NetOperationCompleted
		End If

	Case SendHostRequest
		If (hand >= 0) Then
			' populate the NTP request buffer
			Dim ntpBuf(1 to ntpDataSize) as Byte
			Call MemSet(@ntpBuf, ntpDataSize, 0)
			ntpBuf(1) = &He3	' LI=3, VN=4, Mode=3
			ntpBuf(3) = 6		' poll=6
			ntpBuf(4) = &Hec	' precision=-20

			' send the request to the server
			stat = Net.SendData(hand, ntpBuf, ntpDataSize)
			Debug.Print "Net.SendData() returns "; stat
			state = AwaitingHostReply
			startTick = curTick
		Else
			state = NetOperationCompleted
		End If

	Case AwaitingHostReply
		If (haveData) Then
			Call dspTimeNTP(ntpTime)
			haveData = false
			state = DisconnectFromHost
		ElseIf ((curTick - startTick) > FRC2_replyWaitTime) Then
			Debug.Print "Timed out waiting for host reply"
			state = NetOperationCompleted
		End If

	Case DisconnectFromHost
		If (hand >= 0) Then
			stat = Net.Close(hand)
			Debug.Print "Net.Close() returns "; stat
			hand = -1
		End If
		state = NetOperationCompleted

	End Select
End Sub

Sub myPostInit()
	Debug.Print
	hand = -1
	state = AwaitingWifiConnection
	lastState = state
	lastTick = Register.FRC2_COUNT
End Sub

Sub myInit()
	Call Wifi.SetMode(1)
	Call Wifi.SetConfig(ssid, passwd)
End Sub

'------------------------------------------------------------------------------

'
'' netCallback
'
' This is a callback routine for network operations.
'
Sub netCallback(ByVal cbtype as Byte, ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	Debug.Print "in netCallback(), chan "; handle; " ";
	Select Case cbtype
	Case 0
		' connected to remote host
		Debug.Print "cb: connect"
	Case 1
		' disconnected from remote host
		Debug.Print "cb: disconnect"
	Case 2
		' data received from remote host
		Debug.Print "cb: receive ("; cnt; " bytes)"
#if defined(NTP_CLIENT)
		' the returned data packet should be 48 or more bytes in length
		If (cnt >= ntpDataSize) Then
			Dim nd as NtpData_t
			' extract the NTP data (big-endian order)
			nd.b(4) = data(41)
			nd.b(3) = data(42)
			nd.b(2) = data(43)
			nd.b(1) = data(44)
			ntpTime = nd.d
			haveData = True
		End If
#endif
	Case 3
		' sent data to remote host
		Debug.Print "cb: send"
	Case 4
		' write to remote host completed
		Debug.Print "cb: write finish"
	Case 5
		' reconnected to remote host
		Debug.Print "cb: reconnect"
	Case Else
		' unknown callback type
		Debug.Print "cb: unknown ("; cbtype; ")"
	End Select
End Sub

'------------------------------------------------------------------------------

'
'' dspIP
'
' This routine outputs an IP address in "dotted quad" notation preceded
' by an optional caption.
'
Sub dspIP(ByVal caption as String, ByVal ip as UnsignedLong)
	Dim i as Integer
	Dim ipAddr as IPAddress_t

	ipAddr.w = ip
	Debug.Print caption;
	For i = 1 to 4
		Debug.Print ipAddr.b(i);
		If (i < 4) Then
			Debug.Print ".";
		End If
	Next i
End Sub

'------------------------------------------------------------------------------

'
'' dspTimeNTP
'
' This subroutine displays the NTP time (seconds since 1/1/1900 GMT) and
' also decomposes it into hour, minutes and seconds adjusted for a time zone.
'
Sub dspTimeNTP(ByVal timeNTP as UnsignedLong)
	Const secondsPerMinute as UnsignedLong = 60
	Const secondsPerHour as UnsignedLong = secondsPerMinute * 60
	Const secondsPerDay as UnsignedLong = secondsPerHour * 24
	Const tz as Integer = -7
	Dim seconds as UnsignedLong
	Dim days as UnsignedLong
	Dim hour as UnsignedLong
	Dim minute as UnsignedLong

	Debug.Print "NTP time is "; timeNTP; " seconds since 1/1/1900, GMT"

	' adjust to 1/1/1999
	timeNTP = timeNTP - 3629059200 + (secondsPerHour * CULng(CLng(tz)))
	days = timeNTP \ secondsPerDay
	seconds = timeNTP - (days * secondsPerDay)
	hour = seconds \ secondsPerHour
	seconds = seconds -	(hour * secondsPerHour)
	minute = seconds \ secondsPerMinute
	seconds = seconds - (minute * secondsPerMinute)
	Debug.Print "local time is ";
	If (hour < 10) Then Debug.Print "0";
	Debug.Print hour; ":";
	If (minute < 10) Then Debug.Print "0";
	Debug.Print minute; ":";
	If (seconds < 10) Then Debug.Print "0";
	Debug.Print seconds; " GMT"; tz
End Sub

