'----------------------------------------------------------------------
'
' Copyright 2015 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------

' This example application demonstrates the basics of using an SPI
' Flash File System.  The code mounts a SPIFFS image at the address
' specified by the constant 'flashMountAddr'.  The value &H78000
' used in this example is 32KB from the end of a 512KB Flash chip.
' This allows for a 16KB SPIFFS image (the remaining 16KB is used
' for ESP8266 system parameters).
'
' The Flash area specified by the File.Mount() call must be properly
' initialized.  This can be done by erasing the area using the esp_tool
' utility or a SPIFFS image may be written to it using the same esp_tool
' utility.  An example SPIFFS image is included with this example with
' the name spiffy.bin.

Option ConsoleSpeed 230400

' uncomment this line to perform a write test
'#define SPIFFS_WRITE_TEST

' specify the address at which the SPIFF will be mounted
Const flashMountAddr as UnsignedLong = &H00078000	' 512K - 32K
Const SPIFFS_size as UnsignedLong = 16384

Const SPIFFS_RDONLY as Integer = 8
Const SPIFFS_RDWR as Integer = 8 + 4
Const SPIFFS_WRITE as Integer = 16 + 8 + 4 + 2

Dim state as Integer

Sub Main()
	If (state = 0) Then
		Dim ival as Integer

		Call dspFlash(flashMountAddr, &H500)
		ival = File.Mount(flashMountAddr, SPIFFS_size)
		Debug.Print "File.Mount() returns "; ival
#if defined(SPIFFS_WRITE_TEST)
		' create a file, dump the Flash memory
		Call SPIFFS_writeTest("able.txt")
		Call dspFlash(flashMountAddr, &H500)
#else
		Call SPIFFS_readTest("able.txt")
		Debug.Print
		Call SPIFFS_readTest("baker.txt")
#endif
	End If
	state = state + 1
	If (state > 1000) Then
		state = 1
	End If
End Sub

'
'' SPIFFS_writeTest
'
' This routine creates a file, deleting a same-named file if it exists,
' and writes some data to it.
'
Sub SPIFFS_writeTest(ByVal fname as String)
	Const dataSize as Integer = 16
	Dim buf(1 to dataSize) as Byte
	Dim fh as Integer
	fh = File.Open(fname, SPIFFS_WRITE)
	Debug.Print "File.Open() returns "; fh
	If (fh >= 0) Then
		Dim i as Integer
		Dim lval as Long
		For i = 1 to dataSize
			buf(i) = CByte(i) + &H40
		Next i
		lval = File.Write(fh, buf, SizeOf(buf))
		Debug.Print "File.Write() returns "; lval;
		If (lval < 0) Then
			Debug.Print ", errno = "; File.Error();
		End If
		Debug.Print
		Call File.Close(fh)
	End If
End Sub

'
'' SPIFFS_readTest
'
' This routine attempts to open the named file and, if successful,
' dumps the content in hexadecimal form.
'
Sub SPIFFS_readTest(ByVal fname as String)
	Const dataSize as Integer = 16
	Dim buf(1 to dataSize) as Byte
	Dim fh as Integer

	fh = File.Open(fname, SPIFFS_RDONLY)
	Debug.Print "File.Open() returns "; fh
	If (fh >= 0) Then
		Dim lval as Long

		lval = File.Size(fh)
		If (lval > 0) Then
			Call MemSet(@buf, SizeOf(buf), &H3f)
			lval = File.Read(fh, buf, SizeOf(buf))
			Debug.Print "File.Read() returns "; lval;
			If (lval < 0) Then
				Debug.Print ", errno = "; File.Error();
			End If
			Debug.Print
			Call dspMem(@buf, SizeOf(buf))
		End If
		lval = File.Seek(fh, 0, 2)
		Debug.Print "after reading, File.Seek() returns "; lval
		Call File.Close(fh)
	End If
End Sub

'
'' dspFlash
'
' Display the contents of Flash memory beginning at a given address.
' The output comprises up to 16 hexadecimal bytes per line followed
' by the printable ASCII character equivalents.
'
Sub dspFlash(ByVal addr as UnsignedLong, ByVal cnt as UnsignedInteger)
	Dim i as UnsignedInteger
	Dim b as Byte
	Dim byteCnt as Byte
	Dim dbuf(1 to 16) as Byte

	byteCnt = 0
	For i = 1 to cnt
		If (byteCnt = 0) Then
			Dim result as UnsignedInteger
			Debug.Print CStrHex(addr); "  ";
			result = Flash.Read(addr, dbuf, SizeOf(dbuf))
			addr = addr + CULng(SizeOf(dbuf))
		End If
		byteCnt = byteCnt + 1
		Debug.Print CStrHex(dbuf(byteCnt));
		If (byteCnt = 16) Then
			Dim j as Byte
			Debug.Print "  ";
			For j = 1 to 16
				b = dbuf(j)
				If ((b < &H20) Or (b > &H7e)) Then
					b = &H2e
				End If
				Debug.Print Chr(b);
			Next j
			Debug.Print
			byteCnt = 0
		Else
			Debug.Print " ";
		End If
	Next i
	If (byteCnt <> 0) Then
		Debug.Print
	End If
End Sub

'
'' dspMem
'
' Display the content of memory in hexadecimal form.
'
Sub dspMem(ByVal addr as DataAddress, ByVal cnt as UnsignedInteger)
	Dim b as Byte Based addr
	Dim i as UnsignedInteger
	Dim byteCnt as Byte

	byteCnt = 0
	For i = 1 to cnt
		byteCnt = byteCnt + 1
		Debug.Print CStrHex(b);
		If (byteCnt = 16) Then
			Debug.Print
			byteCnt = 0
		Else
			Debug.Print " ";
		End If
		addr = addr + 1
	Next i
	If (byteCnt <> 0) Then
		Debug.Print
	End If
End Sub
