'----------------------------------------------------------------------
'
' Copyright 2015 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------

' This example demonstrates how to use the WiFi and network routines to implement
' a simple HTTP server.  See the processRequest() subroutine below for the format
' of request that are handled.  The port on which the server listens is given
' by the constant udpPort below.
'
' The Main() routine is structured as a state machine that performs various
' operations in each of several states.  As an aid to understanding the operation,
' much information is output during the process.

Option ConsoleSpeed 230400
Option UserInit myInit
Option UserPostInit myPostInit
Option Include IPAddress_t

' set the SSID and password of the desired WiFi access point
#if Not defined(ssid)
Const ssid as String = "mySSID"
Const passwd as String = "myPasswd"
#endif

' uncomment this line to cause the entire web page to be assembled in a single buffer
#define BUFFERED_OUTPUT

' uncomment this line to add query string support (see processRequest() for details)
#define QUERY_STRING

Const CRLF as String = Chr(&H0d) & Chr(&H0a)

Const httpPort as Integer = 80

Const FRC2_ticksPerSec as UnsignedLong = 5000000

Const WifiConnected as Byte = 5

' define the various states
Enum NetState
	AwaitingWifiConnection
	WaitForConnect
End Enum

' header information for a successful request
Dim normalHeader as ByteVectorData ({
	"HTTP/1.1 200 OK", &H0d, &H0a,
	"Content-Type: text/html; charset=ISO-8859-1", &H0d, &H0a,
	"Cache-Control: no-cache", &H0d, &H0a,
	"Transfer-Encoding: chunked", &H0d, &H0a,
	&H0d, &H0a
})

' partial webserver content
Dim myHTML_head as ByteVectorData ({
	"<html><body>", &H0d, &H0a,
	"<title>ZBasic Web Server</title>", &H0d, &H0a,
	"<center><h2><font color=""#0000FF"">ESP8266</font></h2>", &H0d, &H0a,
	"This is my ZBasic web server page.<br><br>", &H0d, &H0a,
})

Dim myHTML_tail as ByteVectorData ({
	"</center></body></html>", &H0d, &H0a,
	&H0d, &H0a
})

' header information to send when requested content is not found
Dim notFoundHeader as ByteVectorData ({
	"HTTP/1.1 404 Not Found", &H0d, &H0a,
	"Content-Type: text/html; charset=ISO-8859-1", &H0d, &H0a,
	"Cache-Control: no-cache", &H0d, &H0a,
	"Transfer-Encoding: chunked", &H0d, &H0a,
	&H0d, &H0a
})

Dim notFoundHTML as ByteVectorData ({
	"<html><body>", &H0d, &H0a,
	"<title>File Not Found</title>", &H0d, &H0a,
	"<center>The file you requested is not available on this server.", &H0d, &H0a,
	"</center></body></html>", &H0d, &H0a,
	&H0d, &H0a
})

' header information to send when no content is available
Dim noContentHeader as ByteVectorData ({
	"HTTP/1.1 204 No Content", &H0d, &H0a,
	"Content-Length: 0", &H0d, &H0a,
'	"Connection: close", &H0d, &H0a,
	&H0d, &H0a
})

Dim state as NetState
Dim count as Integer
Dim hand as Integer
Dim lastTick as UnsignedLong
Dim lastState as NetState
#if defined(BUFFERED_OUTPUT)
Dim webPageLen as UnsignedInteger
Dim webPage(1 to 1000) as Byte
#endif

Sub Main()
	Dim stat as Integer
	Dim curTick as UnsignedLong
	Dim report as Boolean
	Dim wifiStatus as Byte
	Dim ipAddr as UnsignedLong

	' get the current 200nS tick count and WiFi status
	curTick = Register.FRC2_COUNT
	wifiStatus = Wifi.Status()
	report = False

	' determine if a report should be made
	If (state <> WaitForConnect) Then
		' check for 1 second having elapsed
		If ((curTick - lastTick) > FRC2_ticksPerSec) Then
			report = True
			lastTick = curTick
			count = count + 1
		End If

		' check for a state change
		If (state <> lastState) Then
			report = True
			lastState = state
		End If

		' check for WiFi connection
		If ((state = AwaitingWifiConnection) And (wifiStatus = WifiConnected)) Then
			report = True
		End If
	End If

	' get the current WiFi status, report state information
	If (report) Then
		Debug.Print "count = "; count;
		Debug.Print ", state = "; state;
		Debug.Print ", WiFi status is "; wifiStatus;
		If (hand >= 0) Then
			Debug.Print ", NetStatus() is 0x"; CStrHex(Net.Status(hand));
		End If
		Debug.Print
	End If

	Select Case state
	Case AwaitingWifiConnection
		' see if an AP connection has been established
		If (wifiStatus = WifiConnected) Then
			ipAddr = Wifi.GetIP(0)
			Call dspIP("connected to access point, IP = ", ipAddr)
			Debug.Print

			hand = -1
			If (ipAddr <> 0) Then
				hand = Net.Open()
			End If
			If (hand >= 0) Then
				' set up the network callback, look up host IP address
				stat = Net.SetCallback(hand, @netCallback)
				Debug.Print "Net.SetCallback() returns "; stat

				' begin listening on the HTTP port
				stat = Net.Listen(hand, httpPort, 0)
				Debug.Print "Net.Listen() returns "; stat
			End If
			state = WaitForConnect
		End If

	End Select
End Sub

'------------------------------------------------------------------------------

Sub myPostInit()
	Debug.Print
	hand = -1
	state = AwaitingWifiConnection
	lastState = state
	lastTick = Register.FRC2_COUNT
End Sub

Sub myInit()
	Call Wifi.SetMode(1)
	Call Wifi.SetConfig(ssid, passwd)
End Sub

'------------------------------------------------------------------------------

'
'' netCallback
'
' This is a callback routine for network operations.
'
Sub netCallback(ByVal cbtype as Byte, ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	Dim stat as Integer
	Debug.Print "in netCallback(), chan "; handle; " ";
	Select Case cbtype
	Case 0
		' connected to remote host
		Debug.Print "cb: connect"
	Case 1
		' disconnected from remote host
		Debug.Print "cb: disconnect"
	Case 2
		' data received from remote host
		Debug.Print "cb: receive ("; cnt; " bytes)"
		Call processRequest(handle, data, cnt)
	Case 3
		' sent data to remote host
		Debug.Print "cb: send"
	Case 4
		' write to remote host completed
		Debug.Print "cb: write finish"
	Case 5
		' reconnected to remote host
		Debug.Print "cb: reconnect"
	Case Else
		' unknown callback type
		Debug.Print "cb: unknown ("; cbtype; ")"
	End Select
End Sub

'
'' dspIP
'
' This routine outputs an IP address in "dotted quad" notation preceded
' by an optional caption.
'
Sub dspIP(ByVal caption as String, ByVal ip as UnsignedLong)
	Dim i as Integer
	Dim ipAddr as IPAddress_t

	ipAddr.w = ip
	Debug.Print caption;
	For i = 1 to 4
		Debug.Print ipAddr.b(i);
		If (i < 4) Then
			Debug.Print ".";
		End If
	Next i
End Sub

'------------------------------------------------------------------------------

'
'' dspData
'
' Output a block of data.
'
Sub dspData(ByRef buf() as Byte, ByVal byteCnt as UnsignedInteger)
	Dim i as UnsignedInteger
	Dim needEOL as Boolean

	needEOL = False
	Debug.Print "begin --------------------"
	For i = 1 to byteCnt
		Dim b as Byte
		b = buf(i)
		If (b = &H0a) Then
			Debug.Print
			needEOL = False
		Elseif (b <> &H0d) Then
			Debug.Print Chr(b);
			needEOL = True
		End If
	Next i
	If (needEOL) Then
		Debug.Print
	End If
	Debug.Print "end ----------------------"
End Sub

'------------------------------------------------------------------------------

'
'' processRequest
'
' This subroutine processes a request from an HTTP client.  If the identifier QUERY_STRING
' is defined, the request is checked for a query string with elements as described below.
' A query string is added to a URL by appending a question mark and following it with name-
' value pairs, separated from one another with an ampersand.
'
' The name-value pairs supported are:
'
'      readpin=<pin>
'      setpin=<pin>
'      clrpin=<pin>
'
' where <pin> is either a decimal pin number (1 to 17) or a pin designator like A.0
'
Sub processRequest(ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	If ((handle >= 0) And (cnt > 0)) Then
		Dim stat as Integer
		Call dspData(data, cnt)

		Dim idx as UnsignedInteger
		If checkRequest(data, cnt, idx, "GET /") Or checkRequest(data, cnt, idx, "GET /index.html") Then
#if defined(BUFFERED_OUTPUT)
			Dim dataLen as UnsignedInteger
			Const MaxChunkLen as UnsignedInteger = 7

			' build the entire page in a buffer
			webPageLen = 0
			dataLen = Min(SizeOf(normalHeader), Sizeof(webPage) - webPageLen)
			Call GetProgMem(@normalHeader, webPage(webPageLen + 1), dataLen)
			webPageLen = webPageLen + dataLen

			dataLen = Min(SizeOf(myHTML_head), Sizeof(webPage) - webPageLen - MaxChunkLen)
			Call addChunkSize(handle, dataLen - 2)
			Call GetProgMem(@myHTML_head, webPage(webPageLen + 1), dataLen)
			webPageLen = webPageLen + dataLen

			dataLen = Min(SizeOf(myHTML_tail), Sizeof(webPage) - webPageLen - MaxChunkLen)
			Call addChunkSize(handle, dataLen - 2)
			Call GetProgMem(@myHTML_tail, webPage(webPageLen + 1), dataLen)
			webPageLen = webPageLen + dataLen

			Call addChunkSize(handle, 0)

			stat = Net.SendData(handle, webPage, webPageLen)
#else
			Dim chunkLen as UnsignedInteger

			' send the page in chunks
			stat = Net.SendData(handle, @normalHeader, SizeOf(normalHeader))

			chunkLen = SizeOf(myHTML_head)
			Call addChunkSize(handle, chunkLen - 2)
			stat = Net.SendData(handle, @myHTML_head, chunkLen)

#if defined(QUERY_STRING)
			Call processQueryString(handle, data, cnt, idx)
#endif

			chunkLen = SizeOf(myHTML_tail)
			Call addChunkSize(handle, chunkLen - 2)
			stat = Net.SendData(handle, @myHTML_tail, chunkLen)

			Call addChunkSize(handle, 0)
#endif
		ElseIf checkRequest(data, cnt, idx, "GET /favicon.ico") Then
			stat = Net.SendData(handle, @noContentHeader, SizeOf(noContentHeader))
		Else
#if defined(BUFFERED_OUTPUT)
			Dim dataLen as UnsignedInteger
			Const MaxChunkLen as UnsignedInteger = 7

			' build the entire page in a buffer
			webPageLen = 0
			dataLen = Min(SizeOf(notFoundHeader), Sizeof(webPage) - webPageLen)
			Call GetProgMem(@notFoundHeader, webPage(webPageLen + 1), dataLen)
			webPageLen = webPageLen + dataLen

			dataLen = Min(SizeOf(notFoundHTML), Sizeof(webPage) - webPageLen - MaxChunkLen)
			Call addChunkSize(handle, dataLen - 2)
			Call GetProgMem(@notFoundHTML, webPage(webPageLen + 1), dataLen)
			webPageLen = webPageLen + dataLen

			Call addChunkSize(handle, 0)

			stat = Net.SendData(handle, webPage, webPageLen)
#else
			Dim chunkLen as UnsignedInteger
			stat = Net.SendData(handle, @notFoundHeader, SizeOf(notFoundHeader))
			chunkLen = SizeOf(notFoundHTML)
			Call addChunkSize(handle, chunkLen - 2)
			stat = Net.SendData(handle, @notFoundHTML, chunkLen)
			Call addChunkSize(handle, 0)
#endif
		End If
	End If
End Sub

'------------------------------------------------------------------------------

#if defined(QUERY_STRING)
'
'' processQueryString
'
' Examine a client request string beginning at the specified index to determine
' if a query string is present.  If so, process the query string and emit additional
' response data according to the name/value pairs found.
'
Sub processQueryString(ByVal handle as Integer, ByRef data() as Byte, _
		ByVal cnt as UnsignedInteger, ByVal start as UnsignedInteger)
	Enum QueryElement
		QueryGetPin
		QuerySetPin
		QueryClrPin
		QueryUnknown
	End Enum
	If ((start <= cnt) And (data(start) = Asc("?"))) Then
		start = start + 1
		Do
			Dim pinStartIdx as UnsignedInteger
			Dim elemStartIdx as UnsignedInteger
			Dim idx as UnsignedInteger
			Dim elem as QueryElement
			Dim pin as Byte

			pin = 0
			elemStartIdx = start
			If (isPrefix(data, cnt, start, idx, "getpin=")) Then
				elem = QueryGetPin
			ElseIf (isPrefix(data, cnt, start, idx, "setpin=")) Then
				elem = QuerySetPin
			ElseIf (isPrefix(data, cnt, start, idx, "clrpin=")) Then
				elem = QueryClrPin
			Else
				elem = QueryUnknown
			End If
			pinStartIdx = idx
			If (elem <> QueryUnknown) Then
				pin = getPinNumber(data, cnt, idx)
				If (pin = 0) Then
					elem = QueryUnknown
				End If
			End If

			Dim html as String
			Select Case elem
			Case QueryGetPin
				html = "Pin " & MakeString(@data(pinStartIdx), idx - pinStartIdx) & _
						" = " & CStr(GetPin(pin)) & "<br>" & CRLF
			Case QuerySetPin
				Call PutPin(pin, 1)
				html = "Pin " & MakeString(@data(pinStartIdx), idx - pinStartIdx) & " set<br>" & CRLF
			Case QueryClrPin
				Call PutPin(pin, 0)
				html = "Pin " & MakeString(@data(pinStartIdx), idx - pinStartIdx) & " cleared<br>" & CRLF
			Case Else
				' invalid query string element, locate the next one
				html = ""
				idx = start
				Do While (idx <= cnt)
					If ((data(idx) = Asc(" ")) Or (data(idx) = Asc("&"))) Then
						Exit Do
					End If
					idx = idx + 1
				Loop
			End Select

			' add the HTML code to the
			Dim slen as UnsignedInteger
			slen = CUInt(Len(html))
			If (slen > 0) Then
#if defined(BUFFERED_OUTPUT)
				Dim dataLen as UnsignedInteger
				Const MaxChunkLen as UnsignedInteger = 7

				dataLen = Min(slen, Sizeof(webPage) - webPageLen - MaxChunkLen)
				Call addChunkSize(handle, dataLen - 2)
				Dim i as UnsignedInteger
				For i = 1 to dataLen
					webPage(webPageLen + i) = Asc(html, i)
				Next i
				webPageLen = webPageLen + dataLen
#else
				Dim stat as Integer
				Call addChunkSize(handle, slen - 2)
				stat = Net.SendStr(handle, html)
#endif
			End If

			' check for additional query string elements
			If ((idx <= cnt) And (data(idx) = Asc("&"))) Then
				start = idx + 1
			Else
				Exit Do
			End If
		Loop
	End If
End Sub
#endif

'------------------------------------------------------------------------------

'
'' checkRequest
'
' Determine if a request matches a string.  In order to be deemed to match, the first character
' following the matched string must be either a space, a question mark or CR/LF.
'
Function checkRequest(ByRef data() as Byte, ByVal cnt as UnsignedInteger, _
		ByRef idx as UnsignedInteger, ByVal s as String) as Boolean
	checkRequest = False
	If (isPrefix(data, cnt, 1, idx, s)) Then
		' the string is a prefix of the buffer
		If (idx <= cnt) Then
			' the string is a prefix of the data, check the following character
			Select Case data(idx)
			Case Asc(" "), Asc("?"), &H0d, &H0a
				checkRequest = True
			Case Else
				checkRequest = False
			End Select
		Else
			' exact match
			checkRequest = True
		End If
	End If
End Function

'------------------------------------------------------------------------------

'
'' isPrefix
'
' Determine if the given string is present at the indexed position of the data buffer.
' If so, the updated index refers to the first character beyond the matching string.
'
Function isPrefix(ByRef data() as Byte, ByVal cnt as UnsignedInteger, _
		ByVal start as UnsignedInteger, ByRef idx as UnsignedInteger, _
		ByVal s as String) as Boolean
	Dim slen as UnsignedInteger

	isPrefix = False
	slen = CUInt(Len(s))
	idx = start
	If ((slen + idx) <= cnt) Then
		Dim i as UnsignedInteger

		' see if the data matches the string
		isPrefix = True
		For i = 1 to slen
			If (data(idx) <> Asc(s, i)) Then
				isPrefix = False
				Exit For
			End If
			idx = idx + 1
		Next i
	End If
End Function

'------------------------------------------------------------------------------

'
'' addChunkSize
'
' Add the size of a chunk of data that is about to be transmitted.  The
' size is send as hexadecimal bytes followed by CRLF.  As a special case,
' the chunk size 0 is used as an end marker and is followed by two CRLF pairs.
'
Sub addChunkSize(ByVal handle as Integer, ByVal chunkLen as UnsignedInteger)
	Dim buf(1 to 10) as Byte
	Dim idx as Integer
	Dim firstDigit as Integer
	Dim dataLen as UnsignedInteger
	Dim stat as Integer

	' set the CRLF bytes, determine initial data length
	dataLen = 2
	buf(5) = &H0d
	buf(6) = &H0a
	If (chunkLen = 0) Then
		' for a zero chunk length, add an extra CRLF
		dataLen = 4
		buf(7) = &H0d
		buf(8) = &H0a
	End If

	' generate hex digits, track the leftmost non-zero digit
	firstDigit = 4
	For idx = 4 to 1 Step -1
		Dim val as Byte

		val = LoByte(chunkLen) And &H0f
		If (val <> 0) Then
			firstDigit = idx
		End If
		If (val <= 9) Then
			val = val + Asc("0")
		Else
			val = (val - 10) + Asc("a")
		End If
		buf(idx) = val
		chunkLen = Shr(chunkLen, 4)
	Next idx

#if defined(BUFFERED_OUTPUT)
	' add the chunk information to the page buffer
	dataLen = dataLen + CUInt(5 - firstDigit)
	Call MemCopy(@webPage(webPageLen + 1), @buf(firstDigit), dataLen)
	webPageLen = webPageLen + dataLen
#else
	' send the chunk length information
	stat = Net.SendData(handle, @buf(firstDigit), dataLen + CUInt(5 - firstDigit))
#endif
End Sub

#if defined(QUERY_STRING)
'------------------------------------------------------------------------------

'
'' getPinNumber
'
' Extract a pin designator from a buffer beginning at the given index, returning
' the equivalent pin number or zero if unrecognized or invalid for the target
' device.  A pin number may be specified using the notation P.n (where P is a
' letter specifying the port and n is a digit from 0-7 indicating the bit number
' of the port) or by using the physical pin number.
'
Private Function getPinNumber(ByRef buf() as Byte, ByVal bufLen as UnsignedInteger, ByRef idx as UnsignedInteger) as Byte
	Dim b as Byte
	Dim val as UnsignedInteger
	getPinNumber = 0
	b = toUpperChar(buf(idx))
	If (isCharDigit(b)) Then
		' get the physical pin number, validate
		val = getValue(buf, bufLen, idx)
		getPinNumber = PortBit(LoByte(val))
	ElseIf (isCharAlpha(b) And (buf(idx + 1) = Asc("."))) Then
		' possible port.bit form
		Dim portIdx as Byte
		portIdx = b - Asc("A")
		idx = idx + 2
		b = buf(idx)
		If ((b >= Asc("0")) And (b <= Asc("9"))) Then
			' get the bit number, validate the port.bit
			val = getValue(buf, bufLen, idx)
			getPinNumber = PortBit(portIdx, LoByte(val))
		End If
	End If
End Function

'------------------------------------------------------------------------------

'
'' getValue
'
' Extract a decimal value from a buffer beginning at a particular position.  On
' return, the 'idx' parameter will have been updated to point to the first character
' that was *not* processed.  The value extracted is returned directly.
'
Private Function getValue(ByRef buf() as Byte, ByVal bufLen as UnsignedInteger, ByRef idx as UnsignedInteger) as UnsignedInteger
	getValue = 0
	Do While (idx <= bufLen)
		Dim b as Byte
		b = buf(idx)
		If (isCharDigit(buf(idx))) Then
			getValue = (getValue * 10) + CUInt(b - Asc("0"))
			idx = idx + 1
		Else
			Exit Function
		End If
	Loop
End Function

'------------------------------------------------------------------------------

'
'' isCharDigit
'
' Determine if a character value represents a decimal digit.
'
Private Function isCharDigit(ByVal c as Byte) as Boolean
	isCharDigit = ((c >= Asc("0")) And (c <= Asc("9")))
End Function

'------------------------------------------------------------------------------

'
'' isCharAlpha
'
' Determine if a character value represents an alphabetic character (A-Z or a-z).
'
Private Function isCharAlpha(ByVal c as Byte) as Boolean
	isCharAlpha = (((c >= Asc("A")) And (c <= Asc("Z"))) Or _
			((c >= Asc("a")) And (c <= Asc("z"))))
End Function

'------------------------------------------------------------------------------

'
'' toUpperChar
'
' Convert a character value to upper case if it is lower case.
'
Private Function toUpperChar(ByVal c as Byte) as Byte
	If ((c >= Asc("a")) And (c <= Asc("z"))) Then
		toUpperChar = c - Asc("a") + Asc("A")
	Else
		toUpperChar = c
	End If
End Function

#endif



