'----------------------------------------------------------------------
'
' Copyright 2015 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------

' This example demonstrates how to use the WiFi and network routines to implement
' a simple file-based HTTP server.  This example differs from the ServerHTTP example
' in that pages are loaded from a SPIFFS file system image.
'
' The Main() routine is structured as a state machine that performs various
' operations in each of several states.  As an aid to understanding the operation,
' much information is output during the process.

Option ConsoleSpeed 230400
Option UserInit myInit
Option UserPostInit myPostInit
Option Include IPAddress_t

' set the SSID and password of the desired WiFi access point
#if Not defined(ssid)
Const ssid as String = "mySSID"
Const passwd as String = "myPasswd"
#endif

Const CRLF as String = Chr(&H0d) & Chr(&H0a)

Const httpPort as Integer = 80

Const FRC2_ticksPerSec as UnsignedLong = 5000000
Const FRC2_replyWaitTime as UnsignedLong = FRC2_ticksPerSec * 20

Const WifiConnected as Byte = 5

' specify the address at which the SPIFF will be mounted
Const FlashMountAddr as UnsignedLong = &H00078000	' 512K - 32K
Const SPIFFS_size as UnsignedLong = 16384

Const SPIFFS_RDONLY as Integer = 8

' define the various states
Enum NetState
	AwaitingWifiConnection
	WaitForConnect
End Enum

' header information for a successful request
Dim normalHeader as ByteVectorData ({
	"HTTP/1.1 200 OK", &H0d, &H0a,
	"Content-Type: text/html; charset=ISO-8859-1", &H0d, &H0a,
	"Cache-Control: no-cache", &H0d, &H0a,
	"Transfer-Encoding: chunked", &H0d, &H0a,
	&H0d, &H0a
})

' header information to send when no content is available
Dim noContentHeader as ByteVectorData ({
	"HTTP/1.1 204 No Content", &H0d, &H0a,
	"Content-Length: 0", &H0d, &H0a,
'	"Connection: close", &H0d, &H0a,
	&H0d, &H0a
})

' header information to send when requested content is not found
Dim notFoundHeader as ByteVectorData ({
	"HTTP/1.1 404 Not Found", &H0d, &H0a,
	"Content-Type: text/html; charset=ISO-8859-1", &H0d, &H0a,
	"Cache-Control: no-cache", &H0d, &H0a,
	"Transfer-Encoding: chunked", &H0d, &H0a,
	&H0d, &H0a
})

Dim notFoundHTML as ByteVectorData ({
	"<html><body>", &H0d, &H0a,
	"<title>File Not Found</title>", &H0d, &H0a,
	"<center>The file you requested is not available on this server.", &H0d, &H0a,
	"</center></body></html>", &H0d, &H0a,
	&H0d, &H0a
})

Dim state as NetState
Dim count as Integer
Dim hand as Integer
Dim lastTick as UnsignedLong
Dim lastState as NetState

Sub Main()
	Dim stat as Integer
	Dim curTick as UnsignedLong
	Dim report as Boolean
	Dim wifiStatus as Byte
	Dim ipAddr as UnsignedLong

	' get the current 200nS tick count and WiFi status
	curTick = Register.FRC2_COUNT
	wifiStatus = Wifi.Status()
	report = False

	' determine if a report should be made
	If (state <> WaitForConnect) Then
		' check for 1 second having elapsed
		If ((curTick - lastTick) > FRC2_ticksPerSec) Then
			report = True
			lastTick = curTick
			count = count + 1
		End If

		' check for a state change
		If (state <> lastState) Then
			report = True
			lastState = state
		End If

		' check for WiFi connection
		If ((state = AwaitingWifiConnection) And (wifiStatus = WifiConnected)) Then
			report = True
		End If
	End If

	' get the current WiFi status, report state information
	If (report) Then
		Debug.Print "count = "; count;
		Debug.Print ", state = "; state;
		Debug.Print ", WiFi status is "; wifiStatus;
		If (hand >= 0) Then
			Debug.Print ", NetStatus() is 0x"; CStrHex(Net.Status(hand));
		End If
		Debug.Print
	End If

	Select Case state
	Case AwaitingWifiConnection
		' see if an AP connection has been established
		If (wifiStatus = WifiConnected) Then
			ipAddr = Wifi.GetIP(0)
			Call dspIP("connected to access point, IP = ", ipAddr)
			Debug.Print

			stat = File.Mount(FlashMountAddr, SPIFFS_size)
			Debug.Print "File.Mount() returns "; stat

			hand = -1
			If (ipAddr <> 0) Then
				hand = Net.Open()
			End If
			If (hand >= 0) Then
				' set up the network callback, look up host IP address
				stat = Net.SetCallback(hand, @netCallback)
				Debug.Print "Net.SetCallback() returns "; stat

				' begin listening on the HTTP port
				stat = Net.Listen(hand, httpPort, 0)
				Debug.Print "Net.Listen() returns "; stat
			End If
			state = WaitForConnect
		End If

	End Select
End Sub

'------------------------------------------------------------------------------

Sub myPostInit()
	Debug.Print
	hand = -1
	state = AwaitingWifiConnection
	lastState = state
	lastTick = Register.FRC2_COUNT
End Sub

Sub myInit()
	Call Wifi.SetMode(1)
	Call Wifi.SetConfig(ssid, passwd)
End Sub

'------------------------------------------------------------------------------

'
'' netCallback
'
' This is a callback routine for network operations.
'
Sub netCallback(ByVal cbtype as Byte, ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	Dim stat as Integer
	Debug.Print "in netCallback(), chan "; handle; " ";
	Select Case cbtype
	Case 0
		' connected to remote host
		Debug.Print "cb: connect"
	Case 1
		' disconnected from remote host
		Debug.Print "cb: disconnect"
	Case 2
		' data received from remote host
		Debug.Print "cb: receive ("; cnt; " bytes)"
		Call processRequest(handle, data, cnt)
	Case 3
		' sent data to remote host
		Debug.Print "cb: send"
	Case 4
		' write to remote host completed
		Debug.Print "cb: write finish"
	Case 5
		' reconnected to remote host
		Debug.Print "cb: reconnect"
	Case Else
		' unknown callback type
		Debug.Print "cb: unknown ("; cbtype; ")"
	End Select
End Sub

'------------------------------------------------------------------------------

'
'' dspIP
'
' This routine outputs an IP address in "dotted quad" notation preceded
' by an optional caption.
'
Sub dspIP(ByVal caption as String, ByVal ip as UnsignedLong)
	Dim i as Integer
	Dim ipAddr as IPAddress_t

	ipAddr.w = ip
	Debug.Print caption;
	For i = 1 to 4
		Debug.Print ipAddr.b(i);
		If (i < 4) Then
			Debug.Print ".";
		End If
	Next i
End Sub

'------------------------------------------------------------------------------

'
'' dspData
'
' Output a block of data.
'
Sub dspData(ByRef buf() as Byte, ByVal byteCnt as UnsignedInteger)
	Dim i as UnsignedInteger
	Dim needEOL as Boolean

	needEOL = False
	Debug.Print "begin --------------------"
	For i = 1 to byteCnt
		Dim b as Byte
		b = buf(i)
		If (b = &H0a) Then
			Debug.Print
			needEOL = False
		Elseif (b <> &H0d) Then
			Debug.Print Chr(b);
			needEOL = True
		End If
	Next i
	If (needEOL) Then
		Debug.Print
	End If
	Debug.Print "end ----------------------"
End Sub

'------------------------------------------------------------------------------

'
'' processRequest
'
' This subroutine processes a request from an HTTP client.
'
Sub processRequest(ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	If ((handle >= 0) And (cnt > 0)) Then
		Dim idx as UnsignedInteger
		Dim stat as Integer
		Call dspData(data, cnt)

		If checkRequest(data, cnt, idx, "GET /favicon.ico") Then
			stat = Net.SendData(handle, @noContentHeader, SizeOf(noContentHeader))
		Else
			Dim fname as String
			If checkRequest(data, cnt, idx, "GET /") Then
				fname = "index.html"
			ElseIf ((data(1) = Asc("G")) And (data(2) = Asc("E")) And (data(3) = Asc("T")) And (data(4) = Asc(" "))) Then
				' get the filename, skipping the leading slash (if present)
				idx = 5
				If (data(idx) = Asc("/")) Then
					idx = idx + 1
				End If
				Dim startIdx as UnsignedInteger
				startIdx = idx
				Do While (idx <= cnt)
					If ((data(idx) = Asc(" ")) Or (data(idx) = Asc("?")) Or (data(idx) = &H0d) Or (data(idx) = &H0a)) Then
						Exit Do
					End If
					idx = idx + 1
				Loop
				If ((idx - startIdx) > 0) Then
					fname = MakeString(@data(startIdx), idx - startIdx)
				End If
			End If

			Dim fileFound as Boolean
			fileFound = False
			If (Len(fname) > 0) Then
				Dim fh as Integer

				fh = File.Open(fname, SPIFFS_RDONLY)
				Debug.Print "File.Open() of " & Chr(&H22); fname; Chr(&H22) & " returns "; fh
				If (fh >= 0) Then
					Dim lval as Long

					lval = File.Size(fh)
					If (lval > 0) Then
						' send the page in chunks
						stat = Net.SendData(handle, @normalHeader, SizeOf(normalHeader))

						Call addChunkSize(handle, CUInt(lval))
						Do
							Dim buf(1 to 100) as Byte
							lval = File.Read(fh, buf, SizeOf(buf))
							If (lval <= 0) Then
								Exit Do
							End If
							stat = Net.SendData(handle, buf, lval)
						Loop
						stat = Net.SendStr(handle, CRLF)
						Call addChunkSize(handle, 0)
						fileFound = True
					End If
					Call File.Close(fh)
				End If
			End If

			If (Not fileFound) Then
				Dim chunkLen as UnsignedInteger
				stat = Net.SendData(handle, @notFoundHeader, SizeOf(notFoundHeader))
				chunkLen = SizeOf(notFoundHTML)
				Call addChunkSize(handle, chunkLen - 2)
				stat = Net.SendData(handle, @notFoundHTML, chunkLen)
				Call addChunkSize(handle, 0)
			End If
		End If
	End If
End Sub

'------------------------------------------------------------------------------

'
'' checkRequest
'
' Determine if a request matches a string.  In order to be deemed to match, the first character
' following the matched string must be either a space, a question mark or CR/LF.
'
Function checkRequest(ByRef data() as Byte, ByVal cnt as UnsignedInteger, _
		ByRef idx as UnsignedInteger, ByVal s as String) as Boolean
	checkRequest = False
	If (isPrefix(data, cnt, 1, idx, s)) Then
		' the string is a prefix of the buffer
		If (idx <= cnt) Then
			' the string is a prefix of the data, check the following character
			Select Case data(idx)
			Case Asc(" "), Asc("?"), &H0d, &H0a
				checkRequest = True
			Case Else
				checkRequest = False
			End Select
		Else
			' exact match
			checkRequest = True
		End If
	End If
End Function

'------------------------------------------------------------------------------

'
'' isPrefix
'
' Determine if the given string is present at the indexed position of the data buffer.
' If so, the updated index refers to the first character beyond the matching string.
'
Function isPrefix(ByRef data() as Byte, ByVal cnt as UnsignedInteger, _
		ByVal start as UnsignedInteger, ByRef idx as UnsignedInteger, _
		ByVal s as String) as Boolean
	Dim slen as UnsignedInteger

	isPrefix = False
	slen = CUInt(Len(s))
	idx = start
	If ((slen + idx) <= cnt) Then
		Dim i as UnsignedInteger

		' see if the data matches the string
		isPrefix = True
		For i = 1 to slen
			If (data(idx) <> Asc(s, i)) Then
				isPrefix = False
				Exit For
			End If
			idx = idx + 1
		Next i
	End If
End Function

'------------------------------------------------------------------------------

'
'' addChunkSize
'
' Add the size of a chunk of data that is about to be transmitted.  The
' size is send as hexadecimal bytes followed by CRLF.  As a special case,
' the chunk size 0 is used as an end marker and is followed by two CRLF pairs.
'
Sub addChunkSize(ByVal handle as Integer, ByVal chunkLen as UnsignedInteger)
	Dim buf(1 to 10) as Byte
	Dim idx as Integer
	Dim firstDigit as Integer
	Dim dataLen as UnsignedInteger
	Dim stat as Integer

	' set the CRLF bytes, determine initial data length
	dataLen = 2
	buf(5) = &H0d
	buf(6) = &H0a
	If (chunkLen = 0) Then
		' for a zero chunk length, add an extra CRLF
		dataLen = 4
		buf(7) = &H0d
		buf(8) = &H0a
	End If

	' generate hex digits, track the leftmost non-zero digit
	firstDigit = 4
	For idx = 4 to 1 Step -1
		Dim val as Byte

		val = LoByte(chunkLen) And &H0f
		If (val <> 0) Then
			firstDigit = idx
		End If
		If (val <= 9) Then
			val = val + Asc("0")
		Else
			val = (val - 10) + Asc("a")
		End If
		buf(idx) = val
		chunkLen = Shr(chunkLen, 4)
	Next idx

	' send the chunk length information
	stat = Net.SendData(handle, @buf(firstDigit), dataLen + CUInt(5 - firstDigit))
End Sub




