'----------------------------------------------------------------------
'
' Copyright 2015 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------

' This example demonstrates how to use the WiFi and network routines to implement
' a simple UDP server.  See the processRequest() subroutine below for the format
' of request that are handled.  The port on which the server listens is given
' by the constant udpPort below.
'
' The Main() routine is structured as a state machine that performs various
' operations in each of several states.  As an aid to understanding the operation,
' much information is output during the process.

Option ConsoleSpeed 230400
Option UserInit myInit
Option UserPostInit myPostInit
Option Include IPAddress_t

' set the SSID and password of the desired WiFi access point
#if Not defined(ssid)
Const ssid as String = "mySSID"
Const passwd as String = "myPasswd"
#endif

' set the port number on which the serve will listen
Const udpPort as Integer = 8888

Const FRC2_ticksPerSec as UnsignedLong = 5000000
Const FRC2_replyWaitTime as UnsignedLong = FRC2_ticksPerSec * 20

Const WifiConnected as Byte = 5

' define the various states
Enum NetState
	AwaitingWifiConnection
	WaitForConnect
End Enum

Dim state as NetState
Dim count as Integer
Dim hand as Integer
Dim lastTick as UnsignedLong
Dim lastState as NetState

Sub Main()
	Dim stat as Integer
	Dim curTick as UnsignedLong
	Dim report as Boolean
	Dim wifiStatus as Byte
	Dim ipAddr as UnsignedLong

	' get the current 200nS tick count and WiFi status
	curTick = Register.FRC2_COUNT
	wifiStatus = Wifi.Status()
	report = False

	' determine if a report should be made
	If (state <> WaitForConnect) Then
		' check for 1 second having elapsed
		If ((curTick - lastTick) > FRC2_ticksPerSec) Then
			report = True
			lastTick = curTick
			count = count + 1
		End If

		' check for a state change
		If (state <> lastState) Then
			report = True
			lastState = state
		End If

		' check for WiFi connection
		If ((state = AwaitingWifiConnection) And (wifiStatus = WifiConnected)) Then
			report = True
		End If
	End If

	' get the current WiFi status, report state information
	If (report) Then
		Debug.Print "count = "; count;
		Debug.Print ", state = "; state;
		Debug.Print ", WiFi status is "; wifiStatus;
		If (hand >= 0) Then
			Debug.Print ", NetStatus() is 0x"; CStrHex(Net.Status(hand));
		End If
		Debug.Print
	End If

	Select Case state
	Case AwaitingWifiConnection
		' see if an AP connection has been established
		If (wifiStatus = WifiConnected) Then
			ipAddr = Wifi.GetIP(0)
			Call dspIP("connected to access point, IP = ", ipAddr)
			Debug.Print

			hand = -1
			If (ipAddr <> 0) Then
				hand = Net.Open()
			End If
			If (hand >= 0) Then
				' set up the network callback, look up host IP address
				stat = Net.SetCallback(hand, @netCallback)
				Debug.Print "Net.SetCallback() returns "; stat

				' begin listening on the UDP port
				stat = Net.Listen(hand, udpPort, 1)
				Debug.Print "Net.Listen() returns "; stat
			End If
			state = WaitForConnect
		End If

	End Select
End Sub

'------------------------------------------------------------------------------

Sub myPostInit()
	Debug.Print
	hand = -1
	state = AwaitingWifiConnection
	lastState = state
	lastTick = Register.FRC2_COUNT
End Sub

Sub myInit()
	Call Wifi.SetMode(1)
	Call Wifi.SetConfig(ssid, passwd)
End Sub

'------------------------------------------------------------------------------

'
'' netCallback
'
' This is a callback routine for network operations.
'
Sub netCallback(ByVal cbtype as Byte, ByVal handle as Integer, ByRef data() as Byte, ByVal cnt as UnsignedInteger)
	Dim stat as Integer
	Debug.Print "in netCallback(), chan "; handle; " ";
	Select Case cbtype
	Case 0
		' connected to remote host
		Debug.Print "cb: connect"
	Case 1
		' disconnected from remote host
		Debug.Print "cb: disconnect"
	Case 2
		' data received from remote host
		Debug.Print "cb: receive ("; cnt; " bytes)"
		Call processRequest(handle, data, cnt)
	Case 3
		' sent data to remote host
		Debug.Print "cb: send"
		stat = Net.Close(handle)
	Case 4
		' write to remote host completed
		Debug.Print "cb: write finish"
	Case 5
		' reconnected to remote host
		Debug.Print "cb: reconnect"
	Case Else
		' unknown callback type
		Debug.Print "cb: unknown ("; cbtype; ")"
	End Select
End Sub

'
'' dspIP
'
' This routine outputs an IP address in "dotted quad" notation preceded
' by an optional caption.
'
Sub dspIP(ByVal caption as String, ByVal ip as UnsignedLong)
	Dim i as Integer
	Dim ipAddr as IPAddress_t

	ipAddr.w = ip
	Debug.Print caption;
	For i = 1 to 4
		Debug.Print ipAddr.b(i);
		If (i < 4) Then
			Debug.Print ".";
		End If
	Next i
End Sub

'------------------------------------------------------------------------------

'
'' dspData
'
' Output a block of data.
'
Sub dspData(ByRef buf() as Byte, ByVal byteCnt as UnsignedInteger)
	Dim i as UnsignedInteger
	Dim needEOL as Boolean

	needEOL = False
	Debug.Print "begin --------------------"
	For i = 1 to byteCnt
		Dim b as Byte
		b = buf(i)
		If (b = &H0a) Then
			Debug.Print
			needEOL = False
		Elseif (b <> &H0d) Then
			Debug.Print Chr(b);
			needEOL = True
		End If
	Next i
	If (needEOL) Then
		Debug.Print
	End If
	Debug.Print "end ----------------------"
End Sub

'------------------------------------------------------------------------------

'
'' processRequest
'
' This subroutine processes a request from a UDP client.  The request forms that are
' accepted are as follows:
'
'  W<pin>,<value>
'  R<pin>
'
' Where:
'   <pin>   is a pin number (1-17) or a pin designator such as A.0
'   <value> is either 0 or 1
'
' Note that there is no carriage return/line feed at the end of the request.  The
' presence of *any* additional characters will cause the request to be not
' accepted.
'
' If the request has a recognized and supported form, the reply to the client for
' the W command is "ack" and for the R command it is either "0" or "1".  In all
' other cases the reply is "nak".
'
Sub processRequest(ByVal handle as Integer, ByRef data() as Byte, ByVal byteCnt as UnsignedInteger)
	If ((handle >= 0) And (byteCnt > 0)) Then
		' parse the command in the request packet
		Dim reply as String
		Dim idx as UnsignedInteger
		Dim pin as Byte
		Dim stat as Integer

		Call dspData(data, byteCnt)
		Debug.Print

		reply = "nak"
		idx = 2
		Select Case data(1)
		Case Asc("W")	' write a value to a pin
			' read the pin designator
			pin = getPinNumber(data, byteCnt, idx)
			If ((pin <> 0) And (idx <= byteCnt) And (data(idx) = Asc(","))) Then
				' have valid pin designator followed by comma
				idx = idx + 1

				' validate the desired pin state
				If ((idx = byteCnt) And ((data(idx) = Asc("0")) Or (data(idx) = Asc("1")))) Then
					' set the pin state, indicate success
					Call PutPin(pin, data(idx) - Asc("0"))
					reply = "ack"
				End If
			End If

		Case Asc("R")	' read a pin
			' read the pin designator
			pin = getPinNumber(data, byteCnt, idx)
			If ((pin <> 0) And (idx = byteCnt + 1)) Then
				' set the reply to indicate the state of the pin
				reply = CStr(GetPin(pin))
			End If
		End Select
		Debug.Print "sending reply to the client ["; reply; "]"
		stat = Net.SendStr(handle, reply)
	End If
End Sub

'------------------------------------------------------------------------------

'
'' getPinNumber
'
' Extract a pin designator from a buffer beginning at the given index, returning
' the equivalent pin number or zero if unrecognized or invalid for the target
' device.  A pin number may be specified using the notation P.n (where P is a
' letter specifying the port and n is a digit from 0-7 indicating the bit number
' of the port) or by using the physical pin number.
'
Private Function getPinNumber(ByRef buf() as Byte, ByVal bufLen as UnsignedInteger, ByRef idx as UnsignedInteger) as Byte
	Dim b as Byte
	Dim val as UnsignedInteger
	getPinNumber = 0
	b = toUpperChar(buf(idx))
	If (isCharDigit(b)) Then
		' get the physical pin number, validate
		val = getValue(buf, bufLen, idx)
		getPinNumber = PortBit(LoByte(val))
	ElseIf (isCharAlpha(b) And (buf(idx + 1) = Asc("."))) Then
		' possible port.bit form
		Dim portIdx as Byte
		portIdx = b - Asc("A")
		idx = idx + 2
		b = buf(idx)
		If ((b >= Asc("0")) And (b <= Asc("9"))) Then
			' get the bit number, validate the port.bit
			val = getValue(buf, bufLen, idx)
			getPinNumber = PortBit(portIdx, LoByte(val))
		End If
	End If
End Function

'------------------------------------------------------------------------------

'
'' getValue
'
' Extract a decimal value from a buffer beginning at a particular position.  On
' return, the 'idx' parameter will have been updated to point to the first character
' that was *not* processed.  The value extracted is returned directly.
'
Private Function getValue(ByRef buf() as Byte, ByVal bufLen as UnsignedInteger, ByRef idx as UnsignedInteger) as UnsignedInteger
	getValue = 0
	Do While (idx <= bufLen)
		Dim b as Byte
		b = buf(idx)
		If (isCharDigit(buf(idx))) Then
			getValue = (getValue * 10) + CUInt(b - Asc("0"))
			idx = idx + 1
		Else
			Exit Function
		End If
	Loop
End Function

'------------------------------------------------------------------------------

'
'' isCharDigit
'
' Determine if a character value represents a decimal digit.
'
Private Function isCharDigit(ByVal c as Byte) as Boolean
	isCharDigit = ((c >= Asc("0")) And (c <= Asc("9")))
End Function

'------------------------------------------------------------------------------

'
'' isCharAlpha
'
' Determine if a character value represents an alphabetic character (A-Z or a-z).
'
Private Function isCharAlpha(ByVal c as Byte) as Boolean
	isCharAlpha = (((c >= Asc("A")) And (c <= Asc("Z"))) Or _
			((c >= Asc("a")) And (c <= Asc("z"))))
End Function

'------------------------------------------------------------------------------

'
'' toUpperChar
'
' Convert a character value to upper case if it is lower case.
'
Private Function toUpperChar(ByVal c as Byte) as Byte
	If ((c >= Asc("a")) And (c <= Asc("z"))) Then
		toUpperChar = c - Asc("a") + Asc("A")
	Else
		toUpperChar = c
	End If
End Function


